#ifndef CONSTANT_H
#define CONSTANT_H


#include "petsc.h"


namespace Constant
{


	namespace Time
	{
		extern const PetscScalar& DeltaT;      /**< \brief %time step \f$ \Delta t \f$ */
		extern const PetscInt&    TBeginEq;    /**< \brief restart simulation time */
		extern const PetscInt&    TEndEq;      /**< \brief number of time steps during equilibrium */
		extern const PetscInt&    WriteEveryEq;/**< \brief save every N time steps during equilibrium */
		extern const PetscInt&    PrintEveryEq;/**< \brief print every N time steps during equilibrium */
		extern const PetscInt&    TBegin;      /**< \brief restart simulation time */
		extern const PetscInt&    TEnd;        /**< \brief number of time steps */
		extern const PetscInt&    WriteEvery;  /**< \brief save every N time steps */
		extern const PetscInt&    PrintEvery;  /**< \brief print every N time steps */
	};


	namespace Grid
	{
		extern const PetscInt&    Nx; /**< \brief node number in x-direction \f$ N_x \f$ */
		extern const PetscInt&    Ny; /**< \brief node number in y-direction \f$ N_y \f$ */
		extern const PetscInt&    Nz; /**< \brief node number in z-direction \f$ N_y \f$ */
		extern const PetscScalar& Dx; /**< \brief grid spacing \f$ \Delta x \f$ */
		extern const PetscScalar* L;  /**< \brief domain length \f$ L = \left[ L_x \; L_y \; L_z \right]^\top \f$ */
		extern const PetscScalar* HL; /**< \brief half domain length \f$ HL = \frac12 \left[ L_x \; L_y \; L_z \right]^\top  \f$ */
	}


	namespace Template
	{
		extern const PetscScalar& Thickness; /**< \brief Thickness of the template */
		extern const PetscScalar& Gap;       /**< \brief Gap between base and side */
		extern const PetscInt&    Sides;     /**< \brief Number of sides forming the pyramid */
	}


	namespace Hinge
	{
		extern const PetscScalar& Ec; /**< \brief Elasto-capillary number \f$ Ec \f$ */
		extern const PetscScalar& Damping; /**< \brief Material damping */
	}


	namespace Fluid
	{
		extern const PetscScalar& Xi; /**< \brief length of the A/B diffuse interface \f$ \xi \f$ */
		extern const PetscScalar& Re; /**< \brief Reynolds number \f$ Re \f$ */
		extern const PetscScalar& We; /**< \brief Weber number \f$ We \f$ */
		extern const PetscScalar& Pe; /**< \brief Peclet number \f$ Pe \f$ */
// 		extern const PetscScalar& RhoRatio; /**< \brief fluid B to fluid A density ratio   \f$ \rho_B / \rho_A \f$ */
// 		extern const PetscScalar& EtaRatio; /**< \brief fluid B to fluid A viscosity ratio \f$ \eta_B / \eta_A \f$ */
	}


	namespace Drop
	{
		extern const PetscScalar& Volume; /**< \brief growth coefficient conserving bubble volume */
		extern const PetscScalar& GrowthCoef; /**< \brief growth coefficient conserving bubble volume */
		extern const PetscScalar& ContactAngle; /**< \brief contact angle */
		extern const PetscScalar& IsoPhi; /**< \brief iso value representing drop */
		extern const PetscScalar& SmoothingCoef; /**< \brief curve smoothing factor */
	}


	namespace External
	{
		extern const PetscScalar* Gravity; /** \brief gravity vector */
	}


	namespace Numerical
	{
		/** \brief theta rate
		 *         0   = Explicit,
		 *         0.5 = Crank Nicholson,
		 *         1   = Implicit */
		extern const PetscScalar& ThetaPhi;
	}


	 /** \brief read constants from input file */
	PetscErrorCode ReadInputFile(const std::string &file="Input.in");


}


#endif
