#include "Discretisation3D.h"


//	3D matrix index notation:
//
// 	[	222 221 220 221 222
// 		221 211 210 211 221
// 		220 210 200 210 220
// 		221 211 210 211 221
// 		222 221 220 221 222	]
//
// 		[	221 211 210 211 221
// 			211 111 110 111 211
// 			210 110 100 110 210
// 			211 111 110 111 211
// 			221 211 210 211 221	]
//
// 			[	220 210 200 210 220
// 				210 110 100 110 210
// 				200 100 000 100 200
// 				210 110 100 110 210
// 				220 210 200 210 220	]
//
// 				[	221 211 210 211 221
// 					211 111 110 111 211
// 					210 110 100 110 210
// 					211 111 110 111 211
// 					221 211 210 211 221	]
//
// 					[	222 221 220 221 222
// 						221 211 210 211 221
// 						220 210 200 210 220
// 						221 211 210 211 221
// 						222 221 220 221 222	]


GetDirectionIndices();


namespace
{
	const PetscScalar FOURTH_ORDER_GRAD_27_BY_24_ = 27./24.; /**< \brief Fourth-order gradient coefficient = 27/24 */
	const PetscScalar FOURTH_ORDER_GRAD_01_BY_24_ =  1./24.; /**< \brief Fourth-order gradient coefficient = 1/24 */

	const PetscScalar SECOND_ORDER_LAP_000_ =  -6.; /**< \brief Second-order Laplacian coefficient = -6 */
	const PetscScalar SECOND_ORDER_LAP_100_ =   1.; /**< \brief Second-order Laplacian coefficient = 1 */
	const PetscScalar FOURTH_ORDER_LAP_000_ = -24./6.; /**< \brief Fourth-order Laplacian coefficient) = -24/6 */
	const PetscScalar FOURTH_ORDER_LAP_100_ =   2./6.; /**< \brief Fourth-order Laplacian coefficient = 2/6 */
	const PetscScalar FOURTH_ORDER_LAP_110_ =   1./6.; /**< \brief Fourth-order Laplacian coefficient = 1/6 */
	const PetscScalar FIFTH_ORDER_LAP_000_  =-128./30.; /**< \brief Fifth-order Laplacian coefficient = -128/30 */
	const PetscScalar FIFTH_ORDER_LAP_100_  =  14./30.; /**< \brief Fifth-order Laplacian coefficient = 14/30 */
	const PetscScalar FIFTH_ORDER_LAP_110_  =  3. /30.; /**< \brief Fifth-order Laplacian coefficient) = 3/30 */
	const PetscScalar FIFTH_ORDER_LAP_111_  =  1. /30.; /**< \brief Fifth-order Laplacian coefficient = 1/30 */

	const PetscScalar SECOND_ORDER_BIH_000_ = 42.; /**< \brief Second-order biharmonic coefficient = 42 */
	const PetscScalar SECOND_ORDER_BIH_100_ =-12.; /**< \brief Second-order biharmonic coefficient =-12 */
	const PetscScalar SECOND_ORDER_BIH_110_ =  2.; /**< \brief Second-order biharmonic coefficient =  2 */
	const PetscScalar SECOND_ORDER_BIH_200_ =  1.; /**< \brief Second-order biharmonic coefficient =  1 */
	const PetscScalar FOURTH_ORDER_BIH_000_ = 612./36.; /**< \brief Fourth-order biharmonic coefficient = 612/36 */
	const PetscScalar FOURTH_ORDER_BIH_100_ = -80./36.; /**< \brief Fourth-order biharmonic coefficient = -80/36 */
	const PetscScalar FOURTH_ORDER_BIH_110_ = -36./36.; /**< \brief Fourth-order biharmonic coefficient = -36/36 */
	const PetscScalar FOURTH_ORDER_BIH_111_ =  12./36.; /**< \brief Fourth-order biharmonic coefficient =  12/36 */
	const PetscScalar FOURTH_ORDER_BIH_200_ =   8./36.; /**< \brief Fourth-order biharmonic coefficient =   8/36 */
	const PetscScalar FOURTH_ORDER_BIH_210_ =   4./36.; /**< \brief Fourth-order biharmonic coefficient =   4/36 */
	const PetscScalar FOURTH_ORDER_BIH_211_ =   2./36.; /**< \brief Fourth-order biharmonic coefficient =   2/36 */
	const PetscScalar FOURTH_ORDER_BIH_220_ =   1./36.; /**< \brief Fourth-order biharmonic coefficient =   1/36 */
	const PetscScalar FIFTH_ORDER_BIH_000_ = 17676./900.; /**< \brief Fifth-order biharmonic coefficient = 17676/900 */
	const PetscScalar FIFTH_ORDER_BIH_100_ = -3224./900.; /**< \brief Fifth-order biharmonic coefficient = -3224/900 */
	const PetscScalar FIFTH_ORDER_BIH_110_ =  -284./900.; /**< \brief Fifth-order biharmonic coefficient =  -284/900 */
	const PetscScalar FIFTH_ORDER_BIH_111_ =    -4./900.; /**< \brief Fifth-order biharmonic coefficient =    -4/900 */
	const PetscScalar FIFTH_ORDER_BIH_200_ =   236./900.; /**< \brief Fifth-order biharmonic coefficient =   236/900 */
	const PetscScalar FIFTH_ORDER_BIH_210_ =    96./900.; /**< \brief Fifth-order biharmonic coefficient =    96/900 */
	const PetscScalar FIFTH_ORDER_BIH_211_ =    46./900.; /**< \brief Fifth-order biharmonic coefficient =    46/900 */
	const PetscScalar FIFTH_ORDER_BIH_220_ =    11./900.; /**< \brief Fifth-order biharmonic coefficient =    11/900 */
	const PetscScalar FIFTH_ORDER_BIH_221_ =     6./900.; /**< \brief Fifth-order biharmonic coefficient =     6/900 */
	const PetscScalar FIFTH_ORDER_BIH_222_ =     1./900.; /**< \brief Fifth-order biharmonic coefficient =     1/900 */

	const PetscScalar WENO_13_BY_12_ = 13./12.; /**< \brief Weno coefficient = 13/12 */
	const PetscScalar WENO_01_BY_04_ = 1./4.; /**< \brief Weno coefficient = 1/4 */
	const PetscScalar WENO_01_BY_06_ = 1./6.; /**< \brief Weno coefficient = 1/6 */
	const PetscScalar WENO_02_BY_06_ = 2./6.; /**< \brief Weno coefficient = 2/6 */
	const PetscScalar WENO_05_BY_06_ = 5./6.; /**< \brief Weno coefficient = 5/6 */
	const PetscScalar WENO_07_BY_06_ = 7./6.; /**< \brief Weno coefficient = 7/6 */
	const PetscScalar WENO_11_BY_06_ = 11./6.; /**< \brief Weno coefficient = 11/6 */
	const PetscScalar SMALL_ = 1e-8; /**< \brief Arbitrarily small positive value = 1e-8 */


	/** \brief Left array of first-order qx derivatives
	 *  \param[in]  f  three-dimensional array
	 *  \param[in]  k  node index
	 *  \param[in]  j  node index
	 *  \param[in]  i  node index
	 *  \param[out] qx derivatives \f$ q^{-}_{i=0,..,4} = \partial_x f_i \f$ */
	inline PetscErrorCode WenoLeftQx
	(
		const PetscScalar *const *const *const f,
		PetscInt k,
		PetscInt j,
		PetscInt i,
		PetscScalar qx[5]
	)
	{
		qx[0] = f[k  ][j  ][i-2] - f[k  ][j  ][i-3];
		qx[1] = f[k  ][j  ][i-1] - f[k  ][j  ][i-2];
		qx[2] = f[k  ][j  ][i  ] - f[k  ][j  ][i-1];
		qx[3] = f[k  ][j  ][i+1] - f[k  ][j  ][i  ];
		qx[4] = f[k  ][j  ][i+2] - f[k  ][j  ][i+1];
		return(0);
	}


	/** \brief Left array of first-order qy derivatives
	*  \param[in]  f  three-dimensional array
	*  \param[in]  k  node index
	*  \param[in]  j  node index
	*  \param[in]  i  node index
	*  \param[out] qy derivatives \f$ q^{-}_{i=0,..,4} = \partial_y f_i \f$ */
	inline PetscErrorCode WenoLeftQy
	(
		const PetscScalar *const *const *const f,
		PetscInt k,
		PetscInt j,
		PetscInt i,
		PetscScalar qy[5]
	)
	{
		qy[0] = f[k  ][j-2][i  ] - f[k  ][j-3][i  ];
		qy[1] = f[k  ][j-1][i  ] - f[k  ][j-2][i  ];
		qy[2] = f[k  ][j  ][i  ] - f[k  ][j-1][i  ];
		qy[3] = f[k  ][j+1][i  ] - f[k  ][j  ][i  ];
		qy[4] = f[k  ][j+2][i  ] - f[k  ][j+1][i  ];
		return(0);
	}


	/** \brief Left array of first-order qy derivatives
	*  \param[in]  f  three-dimensional array
	*  \param[in]  k  node index
	*  \param[in]  j  node index
	*  \param[in]  i  node index
	*  \param[out] qz derivatives \f$ q^{-}_{i=0,..,4} = \partial_z f_i \f$ */
	inline PetscErrorCode WenoLeftQz
	(
		const PetscScalar *const *const *const f,
		PetscInt k,
		PetscInt j,
		PetscInt i,
		PetscScalar qz[5]
	)
	{
		qz[0] = f[k-2][j  ][i  ] - f[k-3][j  ][i  ];
		qz[1] = f[k-1][j  ][i  ] - f[k-2][j  ][i  ];
		qz[2] = f[k  ][j  ][i  ] - f[k-1][j  ][i  ];
		qz[3] = f[k+1][j  ][i  ] - f[k  ][j  ][i  ];
		qz[4] = f[k+2][j  ][i  ] - f[k+1][j  ][i  ];
		return(0);
	}


	/** \brief Right array of first-order qx derivatives
	 *  \param[in]  f  three-dimensional array
	 *  \param[in]  k  node index
	 *  \param[in]  j  node index
	 *  \param[in]  i  node index
	 *  \param[out] qx derivatives \f$ q^{+}_{i=0,..,4} = \partial_x f_i \f$ */
	inline PetscErrorCode WenoRightQx
	(
		const PetscScalar *const *const *const f,
		PetscInt k,
		PetscInt j,
		PetscInt i,
		PetscScalar qx[5]
	)
	{
		qx[0] = f[k  ][j  ][i+3] - f[k  ][j  ][i+2];
		qx[1] = f[k  ][j  ][i+2] - f[k  ][j  ][i+1];
		qx[2] = f[k  ][j  ][i+1] - f[k  ][j  ][i  ];
		qx[3] = f[k  ][j  ][i  ] - f[k  ][j  ][i-1];
		qx[4] = f[k  ][j  ][i-1] - f[k  ][j  ][i-2];
		return(0);
	}


	/** \brief Right array of first-order qy derivatives
	 *  \param[in]  f  three-dimensional array
	 *  \param[in]  k  node index
	 *  \param[in]  j  node index
	 *  \param[in]  i  node index
	 *  \param[out] qy derivatives \f$ q^{+}_{i=0,..,4} = \partial_y f_i \f$ */
	inline PetscErrorCode WenoRightQy
	(
		const PetscScalar *const *const *const f,
		PetscInt k,
		PetscInt j,
		PetscInt i,
		PetscScalar qy[5]
	)
	{
		qy[0] = f[k  ][j+3][i  ] - f[k  ][j+2][i  ];
		qy[1] = f[k  ][j+2][i  ] - f[k  ][j+1][i  ];
		qy[2] = f[k  ][j+1][i  ] - f[k  ][j  ][i  ];
		qy[3] = f[k  ][j  ][i  ] - f[k  ][j-1][i  ];
		qy[4] = f[k  ][j-1][i  ] - f[k  ][j-2][i  ];
		return(0);
	}


	/** \brief Right array of first-order qz derivatives
	 *  \param[in]  f  three-dimensional array
	 *  \param[in]  k  node index
	 *  \param[in]  j  node index
	 *  \param[in]  i  node index
	 *  \param[out] qy derivatives \f$ q^{+}_{i=0,..,4} = \partial_z f_i \f$ */
	inline PetscErrorCode WenoRightQz
	(
		const PetscScalar *const *const *const f,
		PetscInt k,
		PetscInt j,
		PetscInt i,
		PetscScalar qy[5]
	)
	{
		qy[0] = f[k+3][j  ][i  ] - f[k+2][j  ][i  ];
		qy[1] = f[k+2][j  ][i  ] - f[k+1][j  ][i  ];
		qy[2] = f[k+1][j  ][i  ] - f[k  ][j  ][i  ];
		qy[3] = f[k  ][j  ][i  ] - f[k-1][j  ][i  ];
		qy[4] = f[k-1][j  ][i  ] - f[k-2][j  ][i  ];
		return(0);
	}


	/** \brief Weighted Weno derivative
	 *  \param[in]  q array of first-order derivatives
	 *  \param[out] f derivative \f$ (\partial_\alpha f)_{\alpha=x,y,z} = \sum w_i \left(\sum \alpha_j q_j \right) \f$ */
	PetscErrorCode WenoWeightedDerivative
	(
		PetscScalar q[5],
		PetscScalar &f
	)
	{
		PetscScalar S0 = WENO_13_BY_12_ * PetscSqr( q[0] - 2.*q[1] + q[2] ) + WENO_01_BY_04_ * PetscSqr(    q[0] - 4.*q[1] + 3.*q[2] );
		PetscScalar S1 = WENO_13_BY_12_ * PetscSqr( q[1] - 2.*q[2] + q[3] ) + WENO_01_BY_04_ * PetscSqr(    q[1]           -    q[3] );
		PetscScalar S2 = WENO_13_BY_12_ * PetscSqr( q[2] - 2.*q[3] + q[4] ) + WENO_01_BY_04_ * PetscSqr( 3.*q[2] - 4.*q[3] +    q[4] );
		PetscScalar omega0 = 0.1 / PetscSqr(1.e-6 + S0);
		PetscScalar omega1 = 0.6 / PetscSqr(1.e-6 + S1);
		PetscScalar omega2 = 0.3 / PetscSqr(1.e-6 + S2);
		PetscScalar omegas = omega0 + omega1 + omega2;
		omega0 /= omegas;
		omega1 /= omegas;
		omega2 /= omegas;
		f =
		(
			  omega0 * (  WENO_02_BY_06_ * q[0] - WENO_07_BY_06_ * q[1] + WENO_11_BY_06_ * q[2] )
			+ omega1 * ( -WENO_01_BY_06_ * q[1] + WENO_05_BY_06_ * q[2] + WENO_02_BY_06_ * q[3] )
			+ omega2 * (  WENO_02_BY_06_ * q[2] + WENO_05_BY_06_ * q[3] - WENO_01_BY_06_ * q[4] )
		);
		return(0);
	}


}


PetscScalar Discretisation3D::SecondOrder::GradX
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		f[k][j][i] - f[k][j][i-1]
	);
}


PetscScalar Discretisation3D::SecondOrder::GradY
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		f[k][j][i] - f[k][j-1][i]
	);
}


PetscScalar Discretisation3D::SecondOrder::GradZ
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		f[k][j][i] - f[k-1][j][i]
	);
}


PetscScalar Discretisation3D::SecondOrder::Grad
(
	const PetscScalar *const *const *const f,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return Discretisation3D::SecondOrder::GradPtr[dim](f,k,j,i);
}


PetscScalar Discretisation3D::FourthOrder::GradX
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  FOURTH_ORDER_GRAD_27_BY_24_ * ( f[k][j][i  ] - f[k][j][i-1] )
		- FOURTH_ORDER_GRAD_01_BY_24_ * ( f[k][j][i+1] - f[k][j][i-2] )
	);
}


PetscScalar Discretisation3D::FourthOrder::GradY
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  FOURTH_ORDER_GRAD_27_BY_24_ * ( f[k][j  ][i] - f[k][j-1][i] )
		- FOURTH_ORDER_GRAD_01_BY_24_ * ( f[k][j+1][i] - f[k][j-2][i] )
	);
}


PetscScalar Discretisation3D::FourthOrder::GradZ
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  FOURTH_ORDER_GRAD_27_BY_24_ * ( f[k  ][j][i] - f[k-1][j][i] )
		- FOURTH_ORDER_GRAD_01_BY_24_ * ( f[k+1][j][i] - f[k-2][j][i] )
	);
}


PetscScalar Discretisation3D::FourthOrder::Grad
(
	const PetscScalar *const *const *const f,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return Discretisation3D::FourthOrder::GradPtr[dim](f,k,j,i);
}


PetscScalar Discretisation3D::SecondOrder::Grad2X
(
	const PetscScalar *const *const *const f,
 	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		( f[k  ][j  ][i+1]
		- f[k  ][j  ][i  ]
		- f[k  ][j  ][i-1]
		+ f[k  ][j  ][i-2] ) * 0.5
	);
}


PetscScalar Discretisation3D::SecondOrder::Grad2Y
(
	const PetscScalar *const *const *const f,
 	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		( f[k  ][j+1][i  ]
		- f[k  ][j  ][i  ]
		- f[k  ][j-1][i  ]
		+ f[k  ][j-2][i  ] ) * 0.5
	);
}


PetscScalar Discretisation3D::SecondOrder::Grad2Z
(
	const PetscScalar *const *const *const f,
 	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		( f[k+1][j  ][i  ]
		- f[k  ][j  ][i  ]
		- f[k-1][j  ][i  ]
		+ f[k-2][j  ][i  ] ) * 0.5
	);
}


PetscScalar Discretisation3D::SecondOrder::Grad2
(
	const PetscScalar *const *const *const f,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return Discretisation3D::SecondOrder::Grad2Ptr[dim](f,k,j,i);
}


PetscScalar Discretisation3D::SecondOrder::Divergent
(
	const PetscScalar *const *const *const fx,
	const PetscScalar *const *const *const fy,
	const PetscScalar *const *const *const fz,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  Discretisation3D::SecondOrder::GradX(fx,k  ,j  ,i+1)
		+ Discretisation3D::SecondOrder::GradY(fy,k  ,j+1,i  )
		+ Discretisation3D::SecondOrder::GradZ(fz,k+1,j  ,i  )
	);
}


PetscScalar Discretisation3D::SecondOrder::Divergent
(
	const PetscScalar *const *const *const *const F,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  Discretisation3D::SecondOrder::GradX(F[0],k  ,j  ,i+1)
		+ Discretisation3D::SecondOrder::GradY(F[1],k  ,j+1,i  )
		+ Discretisation3D::SecondOrder::GradZ(F[2],k+1,j  ,i  )
	);
}


PetscScalar Discretisation3D::SecondOrder::Laplacian
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  SECOND_ORDER_LAP_000_ * ( f[k  ][j  ][i  ] )
		+ SECOND_ORDER_LAP_100_ * ( f[k+1][j  ][i  ]
		                          + f[k-1][j  ][i  ]
		                          + f[k  ][j+1][i  ]
		                          + f[k  ][j-1][i  ]
		                          + f[k  ][j  ][i+1]
		                          + f[k  ][j  ][i-1] )
	);
}


PetscScalar Discretisation3D::FourthOrder::Laplacian
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  FOURTH_ORDER_LAP_000_ * ( f[k  ][j  ][i  ] )
		+ FOURTH_ORDER_LAP_100_ * ( f[k+1][j  ][i  ]
		                          + f[k-1][j  ][i  ]
		                          + f[k  ][j+1][i  ]
		                          + f[k  ][j-1][i  ]
		                          + f[k  ][j  ][i+1]
		                          + f[k  ][j  ][i-1] )
		+ FOURTH_ORDER_LAP_110_ * ( f[k+1][j+1][i  ]
		                          + f[k+1][j-1][i  ]
		                          + f[k-1][j+1][i  ]
		                          + f[k-1][j-1][i  ]
		                          + f[k+1][j  ][i+1]
		                          + f[k+1][j  ][i-1]
		                          + f[k-1][j  ][i+1]
		                          + f[k-1][j  ][i-1]
		                          + f[k  ][j+1][i+1]
		                          + f[k  ][j+1][i-1]
		                          + f[k  ][j-1][i-1]
		                          + f[k  ][j-1][i+1] )
	);
}


PetscScalar Discretisation3D::FifthOrder::Laplacian
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  FIFTH_ORDER_LAP_000_ * ( f[k  ][j  ][i  ] )
		+ FIFTH_ORDER_LAP_100_ * ( f[k+1][j  ][i  ]
		                         + f[k-1][j  ][i  ]
		                         + f[k  ][j+1][i  ]
		                         + f[k  ][j-1][i  ]
		                         + f[k  ][j  ][i+1]
		                         + f[k  ][j  ][i-1] )
		+ FIFTH_ORDER_LAP_110_ * ( f[k+1][j+1][i  ]
		                         + f[k+1][j-1][i  ]
		                         + f[k-1][j+1][i  ]
		                         + f[k-1][j-1][i  ]
		                         + f[k+1][j  ][i+1]
		                         + f[k+1][j  ][i-1]
		                         + f[k-1][j  ][i+1]
		                         + f[k-1][j  ][i-1]
		                         + f[k  ][j+1][i+1]
		                         + f[k  ][j+1][i-1]
		                         + f[k  ][j-1][i-1]
		                         + f[k  ][j-1][i+1] )
		+ FIFTH_ORDER_LAP_111_ * ( f[k+1][j+1][i+1]
		                         + f[k+1][j+1][i-1]
		                         + f[k+1][j-1][i+1]
		                         + f[k+1][j-1][i-1]
		                         + f[k-1][j+1][i+1]
		                         + f[k-1][j+1][i-1]
		                         + f[k-1][j-1][i+1]
		                         + f[k-1][j-1][i-1] )
	);
}


PetscScalar Discretisation3D::SecondOrder::Laplacian
(
	const PetscScalar *const eta,
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  eta[C_] * f[k  ][j  ][i  ]
		+ eta[F_] * f[k+1][j  ][i  ]
		+ eta[B_] * f[k-1][j  ][i  ]
		+ eta[N_] * f[k  ][j+1][i  ]
		+ eta[S_] * f[k  ][j-1][i  ]
		+ eta[E_] * f[k  ][j  ][i+1]
		+ eta[W_] * f[k  ][j  ][i-1]
	);
}


PetscScalar Discretisation3D::SecondOrder::TransposedLaplacianX
(
	const PetscScalar *const eta,
	const PetscScalar *const *const *const *const F,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  eta[F_] * (F[Z_][k+1][j  ][i  ] - F[Z_][k+1][j  ][i-1])
		- eta[B_] * (F[Z_][k  ][j  ][i  ] - F[Z_][k  ][j  ][i-1])
		+ eta[N_] * (F[Y_][k  ][j+1][i  ] - F[Y_][k  ][j+1][i-1])
		- eta[S_] * (F[Y_][k  ][j  ][i  ] - F[Y_][k  ][j  ][i-1])
		+ eta[E_] * (F[X_][k  ][j  ][i+1] - F[X_][k  ][j  ][i  ])
		- eta[W_] * (F[X_][k  ][j  ][i  ] - F[X_][k  ][j  ][i-1])
	);
}


PetscScalar Discretisation3D::SecondOrder::TransposedLaplacianY
(
	const PetscScalar *const eta,
	const PetscScalar *const *const *const *const F,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  eta[F_] * (F[Z_][k+1][j  ][i  ] - F[Z_][k+1][j-1][i  ])
		- eta[B_] * (F[Z_][k  ][j  ][i  ] - F[Z_][k  ][j-1][i  ])
		+ eta[N_] * (F[Y_][k  ][j+1][i  ] - F[Y_][k  ][j  ][i  ])
		- eta[S_] * (F[Y_][k  ][j  ][i  ] - F[Y_][k  ][j-1][i  ])
		+ eta[E_] * (F[X_][k  ][j  ][i+1] - F[X_][k  ][j-1][i+1])
		- eta[W_] * (F[X_][k  ][j  ][i  ] - F[X_][k  ][j-1][i  ])
	);
}


PetscScalar Discretisation3D::SecondOrder::TransposedLaplacianZ
(
	const PetscScalar *const eta,
	const PetscScalar *const *const *const *const F,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  eta[F_] * (F[Z_][k+1][j  ][i  ] - F[Z_][k  ][j  ][i  ])
		- eta[B_] * (F[Z_][k  ][j  ][i  ] - F[Z_][k-1][j  ][i  ])
		+ eta[N_] * (F[Y_][k  ][j+1][i  ] - F[Y_][k-1][j+1][i  ])
		- eta[S_] * (F[Y_][k  ][j  ][i  ] - F[Y_][k-1][j  ][i  ])
		+ eta[E_] * (F[X_][k  ][j  ][i+1] - F[X_][k-1][j  ][i+1])
		- eta[W_] * (F[X_][k  ][j  ][i  ] - F[X_][k-1][j  ][i  ])
	);
}


PetscScalar Discretisation3D::SecondOrder::TransposedLaplacian
(
	const PetscScalar *const eta,
	const PetscScalar *const *const *const *const F,
	PetscInt dim,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return Discretisation3D::SecondOrder::TransposedLaplacianPtr[dim](eta,F,k,j,i);
}


PetscScalar Discretisation3D::SecondOrder::Biharmonic
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return
	(
		  SECOND_ORDER_BIH_000_ * ( f[k  ][j  ][i  ] )
		+ SECOND_ORDER_BIH_100_ * ( f[k+1][j  ][i  ]
		                          + f[k-1][j  ][i  ]
		                          + f[k  ][j+1][i  ]
		                          + f[k  ][j-1][i  ]
		                          + f[k  ][j  ][i+1]
		                          + f[k  ][j  ][i-1] )
		+ SECOND_ORDER_BIH_110_ * ( f[k+1][j+1][i  ]
		                          + f[k+1][j-1][i  ]
		                          + f[k-1][j+1][i  ]
		                          + f[k-1][j-1][i  ]
		                          + f[k+1][j  ][i+1]
		                          + f[k+1][j  ][i-1]
		                          + f[k-1][j  ][i+1]
		                          + f[k-1][j  ][i-1]
		                          + f[k  ][j+1][i+1]
		                          + f[k  ][j+1][i-1]
		                          + f[k  ][j-1][i-1]
		                          + f[k  ][j-1][i+1] )
		+ SECOND_ORDER_BIH_200_ * ( f[k+2][j  ][i  ]
		                          + f[k-2][j  ][i  ]
		                          + f[k  ][j+2][i  ]
		                          + f[k  ][j-2][i  ]
		                          + f[k  ][j  ][i+2]
		                          + f[k  ][j  ][i-2] )
	);
}


PetscScalar Discretisation3D::FourthOrder::Biharmonic
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return 0;
}


PetscScalar Discretisation3D::FifthOrder::Biharmonic
(
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	return 0;
}


PetscScalar Discretisation3D::FirstOrderUpwind::UGrad
(
	const PetscScalar *const uDir,
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	PetscScalar fx,fy,fz;
	if      (uDir[0] < -SMALL_) { fx = ( f[k  ][j  ][i+1] - f[k  ][j  ][i  ] ); }
	else if (uDir[0] >  SMALL_) { fx = ( f[k  ][j  ][i  ] - f[k  ][j  ][i-1] ); }
	else    { fx = 0.; }
	if      (uDir[1] < -SMALL_) { fy = ( f[k  ][j+1][i  ] - f[k  ][j  ][i  ] ); }
	else if (uDir[1] >  SMALL_) { fy = ( f[k  ][j  ][i  ] - f[k  ][j-1][i  ] ); }
	else    { fy = 0.; }
	if      (uDir[2] < -SMALL_) { fz = ( f[k+1][j  ][i  ] - f[k  ][j  ][i  ] ); }
	else if (uDir[2] >  SMALL_) { fz = ( f[k  ][j  ][i  ] - f[k-1][j  ][i  ] ); }
	else    { fz = 0.; }
	return
	(
		  uDir[0] * fx
		+ uDir[1] * fy
		+ uDir[2] * fz
	);
}


PetscScalar Discretisation3D::SecondOrderUpwind::UGrad
(
	const PetscScalar *const uDir,
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	PetscScalar fx,fy, fz;
	if      (uDir[0] < -SMALL_) { fx = (-0.5 * f[k  ][j  ][i+2] + 2. * f[k  ][j  ][i+1] - 1.5 * f[k  ][j  ][i  ] ); }
	else if (uDir[0] >  SMALL_) { fx = ( 1.5 * f[k  ][j  ][i  ] - 2. * f[k  ][j  ][i-1] + 0.5 * f[k  ][j  ][i-2] ); }
	else    { fx = 0.; }
	if      (uDir[1] < -SMALL_) { fy = (-0.5 * f[k  ][j+2][i  ] + 2. * f[k  ][j+1][i  ] - 1.5 * f[k  ][j  ][i  ] ); }
	else if (uDir[1] >  SMALL_) { fy = ( 1.5 * f[k  ][j  ][i  ] - 2. * f[k  ][j-1][i  ] + 0.5 * f[k  ][j-2][i  ] ); }
	else    { fy = 0.; }
	if      (uDir[2] < -SMALL_) { fz = (-0.5 * f[k+2][j  ][i  ] + 2. * f[k+1][j  ][i  ] - 1.5 * f[k  ][j  ][i  ] ); }
	else if (uDir[2] >  SMALL_) { fz = ( 1.5 * f[k  ][j  ][i  ] - 2. * f[k-1][j  ][i  ] + 0.5 * f[k-2][j  ][i  ] ); }
	else    { fz = 0.; }

	return
	(
		  uDir[0] * fx
		+ uDir[1] * fy
		+ uDir[2] * fz
	);
}


PetscScalar Discretisation3D::Weno::UGrad
(
	const PetscScalar *const uDir,
	const PetscScalar *const *const *const f,
	PetscInt k,
	PetscInt j,
	PetscInt i
)
{
	PetscScalar q[5];
	PetscScalar fx,fy,fz;
	if      (uDir[0] < -SMALL_){ WenoRightQx(f,k,j,i,q); WenoWeightedDerivative(q,fx); }
	else if (uDir[0] >  SMALL_){ WenoLeftQx (f,k,j,i,q); WenoWeightedDerivative(q,fx); }
	else    { fx = 0.; }
	if      (uDir[1] < -SMALL_){ WenoRightQy(f,k,j,i,q); WenoWeightedDerivative(q,fy); }
	else if (uDir[1] >  SMALL_){ WenoLeftQy (f,k,j,i,q); WenoWeightedDerivative(q,fy); }
	else    { fy = 0.; }
	if      (uDir[2] < -SMALL_){ WenoRightQz(f,k,j,i,q); WenoWeightedDerivative(q,fz); }
	else if (uDir[2] >  SMALL_){ WenoLeftQz (f,k,j,i,q); WenoWeightedDerivative(q,fz); }
	else    { fz = 0.; }
	return
	(
		  uDir[0] * fx
		+ uDir[1] * fy
		+ uDir[2] * fz
	);
}
