#include "EigenWrapper.h"

EigenWrapper::EigenQuat EigenWrapper::AngleAxisToQuat
(
	const PetscScalar& angle,
	const EigenVec3& axis
)
{
	return EigenQuat
	(
		EigenAngleAxis
		(
			angle,
			axis.normalized()
		)
	).normalized();
}

template<typename T>
PetscErrorCode EigenWrapper::WriteInASCIIFormat
(
	const T& obj,
	const std::string &name,
	PetscInt t
)
{
	if (obj.size() > 0)
	{
		char fname[1000];
		if (t<0)
		{
			sprintf(fname,"%s.txt",name.c_str());
		}
		else
		{
			sprintf(fname,"%s_t%.*d.txt",name.c_str(),5,t);
		}
		std::cout << "  Write " << fname << std::endl;
		std::ofstream file(fname);
		if (obj.cols() > 1)
		{
			//Matrix
			Eigen::IOFormat fmt
			(
				8,    // precision
				0,    // flag
				"\t", // coeff separator
				"\n"  // row separator
			);
			file << obj.format(fmt);
		}
		else
		{
			//Vector
			file << obj.transpose();
		}
		file << std::endl;
		file.close();
	}
	return(0);
}

template PetscErrorCode EigenWrapper::WriteInASCIIFormat<EigenWrapper::EigenVec3>(const EigenWrapper::EigenVec3&,const std::string&,PetscInt);
template PetscErrorCode EigenWrapper::WriteInASCIIFormat<EigenWrapper::EigenVec>(const EigenWrapper::EigenVec&,const std::string&,PetscInt);
template PetscErrorCode EigenWrapper::WriteInASCIIFormat<EigenWrapper::EigenMat>(const EigenWrapper::EigenMat&,const std::string&,PetscInt);
template PetscErrorCode EigenWrapper::WriteInASCIIFormat<Eigen::Map<const EigenWrapper::EigenVec>>(const Eigen::Map<const EigenWrapper::EigenVec>&,const std::string&,PetscInt);
template PetscErrorCode EigenWrapper::WriteInASCIIFormat<Eigen::Map<const EigenWrapper::EigenMat>>(const Eigen::Map<const EigenWrapper::EigenMat>&,const std::string&,PetscInt);

template<typename T>
T EigenWrapper::MPIReduce
(
	const T& local
)
{
	T global(local.size());
	MPI_Allreduce
	(
		local.data(),
		global.data(),
		local.size(),
		MPIU_SCALAR,
		MPIU_SUM,
		PETSC_COMM_WORLD
	);
	return global;
}

template EigenWrapper::EigenVec  EigenWrapper::MPIReduce<EigenWrapper::EigenVec >(const EigenWrapper::EigenVec &);
template EigenWrapper::EigenVec3 EigenWrapper::MPIReduce<EigenWrapper::EigenVec3>(const EigenWrapper::EigenVec3&);
