#ifndef EIGENWRAPPER_H
#define EIGENWRAPPER_H

// #define EIGEN_MATRIXBASE_PLUGIN "/home/gl5972/Documents/microorigami/src/Tmp.h"
// #define EIGEN_QUATERNIONBASE_PLUGIN "/home/gl5972/Documents/microorigami/src/Tmp.h"
#include <Eigen/Dense>
#include <petsc.h>
#include <iostream>
#include <fstream>


namespace EigenWrapper
{
	typedef Eigen::Matrix<PetscScalar,Eigen::Dynamic,1> EigenVec; /**< \brief Eigen scalar vector */
	typedef Eigen::Matrix<PetscScalar,2,1> EigenVec2; /**< \brief Eigen 3D-vector */
	typedef Eigen::Matrix<PetscScalar,3,1> EigenVec3; /**< \brief Eigen 3D-vector */
	typedef Eigen::Matrix<PetscScalar,4,1> EigenVec4; /**< \brief Eigen 4D-vector */
	typedef Eigen::Matrix<PetscScalar,5,1> EigenVec5; /**< \brief Eigen 5D-vector */
	typedef Eigen::Matrix<PetscScalar,6,1> EigenVec6; /**< \brief Eigen 6D-vector */
	typedef Eigen::Matrix<int,Eigen::Dynamic,1> EigenVecI; /**< \brief Eigen vector of integers */
	typedef Eigen::Matrix<bool,Eigen::Dynamic,1> EigenVecB; /**< \brief Eigen vector of booleans */

	/** \brief Eigen quaternion.
	 *         Internally the coefficients (i.e. function coeffs() ) are stored in the following order: [x,y,z,w]^T.
	 *         A quaternion is however constructed in the following order [w,x,y,z]^T */
	typedef Eigen::Quaternion<PetscScalar> EigenQuat;

	typedef Eigen::Matrix<PetscScalar,Eigen::Dynamic,Eigen::Dynamic> EigenMat; /**< \brief Eigen scalar matrix */
	typedef Eigen::Matrix<PetscScalar,Eigen::Dynamic,3> EigenMatX3;
	typedef Eigen::Matrix<PetscScalar,2,2> EigenMat2; /**< \brief Eigen 2x2 matrix */
	typedef Eigen::Matrix<PetscScalar,3,3> EigenMat3; /**< \brief Eigen 3x3 matrix */
	typedef Eigen::Matrix<PetscScalar,4,4> EigenMat4; /**< \brief Eigen 4x4 matrix */
	typedef Eigen::Matrix<PetscScalar,5,5> EigenMat5; /**< \brief Eigen 5x5 matrix */
	typedef Eigen::Matrix<PetscScalar,6,6> EigenMat6; /**< \brief Eigen 6x6 matrix */
	typedef Eigen::Matrix<int,Eigen::Dynamic,Eigen::Dynamic> EigenMatI;  /**< \brief Eigen matrix of integers */
	typedef Eigen::AngleAxis<PetscScalar> EigenAngleAxis; /**< \brief Angle Axis object */


	/** \brief Create normalized quaternion given rotation angle and axis */
	EigenQuat AngleAxisToQuat(const PetscScalar& angle,const EigenVec3& axis);

	/** \brief Write Eigen object in ascii format
	 *  \param[in] obj  eigen object
	 *  \param[in] name name
	 *  \param[in] t    time */
	template<typename T>
	PetscErrorCode WriteInASCIIFormat
	(
		const T &obj,
		const std::string &name,
		PetscInt t
	);

	/** \brief Sum vectors from all processes and distributes the result back to all processe
	 *  \param[in] local local eigen vector
	 *  \note only works for scalars. For integers and booleans, it will likely fail
	 *  \return global vector */
	template<typename T>
	T MPIReduce
	(
		const T &local
	);

// 	const EigenVec3 EX_(1,0,0); /**< \brief unit x-vector (1 0 0) */
// 	const EigenVec3 EY_(0,1,0); /**< \brief unit y-vector (0 1 0) */
// 	const EigenVec3 EZ_(0,0,1); /**< \brief unit z-vector (0 0 1) */
}


#endif
