#include "Hinge.h"

static char help[] = "Export hinge.dat to hinge.txt\n\n";

PetscErrorCode PrintUsage()
{
	std::cout
		<< "  USAGE:" << std::endl
		<< "  -h           print usage" << std::endl
		<< "  -f           hinge" << std::endl
		<< "  -t t0:t1:dt  start, end, every" << std::endl
		<< "  Export hinge.dat to hinge.txt" << std::endl;
	return(0);
}

PetscErrorCode GetTimes
(
	PetscInt &t0,
	PetscInt &t1,
	PetscInt &dt
)
{
	PetscBool flg;
	char str[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-t",str,sizeof(str),&flg);
	if (!flg)
	{
		//SETERRQ(PETSC_COMM_WORLD,1,"Pass with -t t0:t1:dt");
		t0 = -1;
		t1 =  0;
		dt =  1;
		return(0);
	}
	char *p = strtok(str,":"); t0 = atoi(p);
	p = strtok(NULL,":"); t1 = (p == NULL) ? t0+1 : atoi(p);
	p = strtok(NULL,":"); dt = (p == NULL) ? 1 : atoi(p);
	return(0);
}

PetscInt main(int argc,char** args)
{
	PetscInitialize(&argc,&args,(char*)0,help);
	PetscBool usage(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-h",&usage,NULL);
	if (usage)
	{
		PrintUsage();
		PetscEnd();
	}
	char hinge[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-f",hinge,sizeof(hinge),NULL);
	PetscInt t0,t1,dt; GetTimes(t0,t1,dt);
	Hinge h(hinge);
	for (PetscInt t=t0; t<t1; t+=dt)
	{
		h.readInBinaryFormat(t);
		h.writeInASCIIFormat(t);
	}
	PetscFinalize();
	return(0);
}
