#include "Polyline.h"

static char help[] = "Export single polyline to human readable format\n\n";

PetscErrorCode PrintUsage()
{
	std::cout
		<< "  USAGE:" << std::endl
		<< "  -h   print usage" << std::endl
		<< "  -f   polyline.dat (binary format)" << std::endl
		<< "  -eps distance used in smoothing operation" << std::endl
		<< "  -s   smoothing factor" << std::endl
		<< "  -t   times (start:end:every)" << std::endl
		<< "  Export single polyline to human readable format" << std::endl
	;
	return(0);
}

PetscErrorCode GetTimes
(
	PetscInt &t0,
	PetscInt &t1,
	PetscInt &dt
)
{
	PetscBool flg;
	char str[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-t",str,sizeof(str),&flg);
	if (!flg)
	{
		//SETERRQ(PETSC_COMM_WORLD,1,"Pass with -t t0:t1:dt");
		t0 = -1;
		t1 =  0;
		dt =  1;
		return(0);
	}
	char *p = strtok(str,":"); t0 = atoi(p);
	p = strtok(NULL,":"); t1 = (p == NULL) ? t0+1 : atoi(p);
	p = strtok(NULL,":"); dt = (p == NULL) ? 1 : atoi(p);
	return(0);
}

PetscInt main(int argc,char** args)
{
	PetscInitialize(&argc,&args,(char*)0,help);
	PetscBool usage(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-h",&usage,NULL);
	if (usage)
	{
		PrintUsage();
		PetscEnd();
	}
	char pname[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-f",pname,sizeof(pname),NULL);
	PetscScalar eps(1); PetscOptionsGetScalar(NULL,NULL,"-eps", &eps,NULL);
	PetscScalar s(-1); PetscOptionsGetScalar(NULL,NULL,"-s", &s,NULL);
	PetscInt t0,t1,dt; GetTimes(t0,t1,dt);
	Polyline pl;
	for (PetscInt t=t0; t<t1; t+=dt)
	{
		pl.readInBinaryFormat(pname,t);
		if (s>0)
		{
			pl.smooth(eps,s);
		}
		pl.writeInAsciiFormat(t);
	}
	PetscFinalize();
	return(0);
}
