#include <iostream>
#include <fstream>
#include "PetscWrapper.h"
#include "EigenWrapper.h"

static char help[] = "Export scalar.dat to scalar.txt\n\n";

PetscErrorCode PrintUsage()
{
	std::cout
		<< "  USAGE:" << std::endl
		<< "  -h           print usage" << std::endl
		<< "  -f           scalar" << std::endl
		<< "  -n           array size" << std::endl
		<< "  -t t0:t1:dt  start, end, every" << std::endl
		<< "  Export scalar.dat to scalar.txt" << std::endl;
	return(0);
}

PetscErrorCode GetTimes
(
	PetscInt &t0,
	PetscInt &t1,
	PetscInt &dt
)
{
	PetscBool flg;
	char str[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-t",str,sizeof(str),&flg);
	if (!flg)
	{
		//SETERRQ(PETSC_COMM_WORLD,1,"Pass with -t t0:t1:dt");
		t0 = -1;
		t1 =  0;
		dt =  1;
		return(0);
	}
	char *p = strtok(str,":"); t0 = atoi(p);
	p = strtok(NULL,":"); t1 = (p == NULL) ? t0+1 : atoi(p);
	p = strtok(NULL,":"); dt = (p == NULL) ? 1 : atoi(p);
	return(0);
}

typedef std::vector<PetscScalar> Scalars;

PetscInt main(int argc,char** args)
{
	PetscInitialize(&argc,&args,(char*)0,help);
	PetscBool usage(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-h",&usage,NULL);
	if (usage)
	{
		PrintUsage();
		PetscEnd();
	}
	char scalar[PETSC_MAX_PATH_LEN]; PetscOptionsGetString(NULL,NULL,"-f",scalar,sizeof(scalar),NULL);
	PetscInt n(1); PetscOptionsGetInt(NULL,NULL,"-n",&n,NULL);
	PetscInt t0,t1,dt; GetTimes(t0,t1,dt);
	Scalars scalars(n);
	std::vector<Scalars> scalarss;
	for (PetscInt t=t0; t<t1; t+=dt)
	{
		if (n>1)
		{
			PetscWrapper::ReadScalarInBinaryFormat(scalars.data(),std::string(scalar),t,scalars.size());
			scalarss.push_back(scalars);
		}
		else
		{
			PetscWrapper::ReadScalarInBinaryFormat(scalars[0],std::string(scalar),t);
			scalarss.push_back(scalars);
		}

	}
	std::string fname = std::string(scalar)+".txt";
	std::cout << "  Write " + fname << std::endl;
	std::ofstream file(fname);
	for (const auto& scalarsi:scalarss)
	{
		for (const auto& scalari:scalarsi)
		{
			file << scalari << " ";
		}
		file << std::endl;
	}
	file.close();
	PetscFinalize();
	return(0);
}
