#include "Field3D.h"


namespace
{
	typedef ScalarField ScFi;
	typedef VectorField VeFi;
}


PetscErrorCode ScalarField::create
(
	DM da,
	const std::string &name
)
{
	if (da)
	{
		PetscObjectReference((PetscObject)da);
		da_ = da;
		PetscWrapper::CreateGlobalPetscVector(da_,g_,name);
		PetscWrapper::CreateLocalPetscVector(da_,l_);
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: dynamic array is undefined\n");
		PetscEnd();
	}
	return(0);
}


PetscErrorCode VectorField::create
(
	DM da,
	const std::string &name,
	PetscInt n
)
{
	if (da)
	{
		PetscObjectReference((PetscObject)da);
		da_ = da;
		n_ = n;
		PetscWrapper::CreateGlobalPetscVector(da_,g_,name,n_);
		PetscWrapper::CreateLocalPetscVector (da_,l_,n_);
		PetscMalloc1(n_,&a_);
	}
	else
	{
		PetscPrintf(PETSC_COMM_WORLD,"  Error: dynamic array is undefined\n");
		PetscEnd();
	}
	return(0);
}


PetscErrorCode VectorField::create3D
(
	DM da,
	const std::string &name
)
{
	this->create(da,name,3);
	return(0);
}


PetscErrorCode ScalarField::destroy()
{
// 	PetscObjectDereference((PetscObject)da);
	DMDestroy(&da_);
	PetscWrapper::DestroyPetscVector(g_);
	PetscWrapper::DestroyPetscVector(l_);
	return(0);
}


PetscErrorCode VectorField::destroy()
{
// 	PetscObjectDereference((PetscObject)da);
	DMDestroy(&da_);
	PetscWrapper::DestroyPetscVector(g_,n_);
	PetscWrapper::DestroyPetscVector(l_,n_);
	PetscFree(a_);
	return(0);
}


PetscErrorCode ScalarField::check()
{
	PetscWrapper::CheckPetscVector(g_);
	return(0);
}


PetscErrorCode VectorField::check()
{
	PetscWrapper::CheckPetscVector(g_,n_);
	return(0);
}


PetscErrorCode ScalarField::writeInBinaryFormat
(
	PetscInt t
) const
{
	PetscWrapper::WritePetscVectorInBinaryFormat(g_,t);
	return(0);
}


PetscErrorCode VectorField::writeInBinaryFormat
(
	PetscInt t
) const
{
	PetscWrapper::WritePetscVectorInBinaryFormat(g_,t,n_);
	return(0);
}


PetscErrorCode ScalarField::writeInVtsFormat
(
	PetscInt t
) const
{
	PetscWrapper::WritePetscVectorInVtsFormat(g_,t);
	return(0);
}


PetscErrorCode VectorField::writeInVtsFormat
(
	PetscInt t
) const
{
	PetscWrapper::WritePetscVectorInVtsFormat(g_,t,n_);
	return(0);
}


PetscErrorCode VectorField::writeInVtsFormatWithXYZSubscripts
(
	PetscInt t
) const
{
	PetscWrapper::WritePetscVectorInVtsFormatWithXYZSubscripts(g_,t,n_);
	return(0);
}


PetscErrorCode ScalarField::readInBinaryFormat
(
	PetscInt t
)
{
	PetscWrapper::ReadPetscVectorInBinaryFormat(g_,t);
	return(0);
}


PetscErrorCode VectorField::readInBinaryFormat
(
	PetscInt t
)
{
	PetscWrapper::ReadPetscVectorInBinaryFormat(g_,t,n_);
	return(0);
}


PetscErrorCode ScalarField::getArray()
{
	PetscWrapper::GetPetscArray(da_,g_,a_);
	return(0);
}


PetscErrorCode VectorField::getArray()
{
	PetscWrapper::GetPetscArray(da_,g_,a_,n_);
	return(0);
}


PetscErrorCode ScalarField::restoreArray()
{
	PetscWrapper::RestorePetscArray(da_,g_,a_);
	return(0);
}


PetscErrorCode VectorField::restoreArray()
{
	PetscWrapper::RestorePetscArray(da_,g_,a_,n_);
	return(0);
}


PetscErrorCode ScalarField::getGhostedArray()
{
	PetscWrapper::GetGhostedPetscArray(da_,g_,l_,a_);
	return(0);
}


PetscErrorCode VectorField::getGhostedArray()
{
	PetscWrapper::GetGhostedPetscArray(da_,g_,l_,a_,n_);
	return(0);
}


PetscErrorCode ScalarField::restoreGhostedArray()
{
	PetscWrapper::RestoreGhostedPetscArray(da_,l_,a_);
	return(0);
}


PetscErrorCode VectorField::restoreGhostedArray()
{
	PetscWrapper::RestoreGhostedPetscArray(da_,l_,a_,n_);
	return(0);
}


PetscErrorCode ScalarField::flattenMarchingCubeArrayTo
(
	std::vector<PetscScalar>& v,
	PetscInt &nx,
	PetscInt &ny,
	PetscInt &nz
)
{
	PetscWrapper::FlattenMarchingCubeArrayTo(da_,g_,l_,a_,v,nx,ny,nz);
	return(0);
}


PetscErrorCode ScalarField::copyTo
(
	ScalarField &to
)
{
	PetscWrapper::CopyPetscVectorTo(this->g_,to.g_);
	return(0);
}


PetscErrorCode VectorField::copyTo
(
	VectorField &to
)
{
	PetscWrapper::CopyPetscVectorTo(this->g_,to.g_,to.n_);
	return(0);
}



PetscErrorCode ScalarField::set
(
	PetscScalar s
)
{
	PetscWrapper::SetPetscVector(g_,s);
	return(0);
}


PetscErrorCode VectorField::set
(
	PetscScalar s
)
{
	PetscWrapper::SetPetscVector(g_,s,n_);
	return(0);
}


PetscErrorCode VectorField::set
(
	const PetscScalar *const S
)
{
	PetscWrapper::SetPetscVector(g_,S,n_);
	return(0);
}


PetscErrorCode ScalarField::zero()
{
	PetscWrapper::ZeroPetscVector(g_);
	return(0);
}


PetscErrorCode VectorField::zero()
{
	PetscWrapper::ZeroPetscVector(g_,n_);
	return(0);
}


PetscErrorCode ScalarField::sum
(
	PetscScalar &s
)
{
	PetscWrapper::ComputeSumOfPetscVectorElements(g_,s);
	return(0);
}


PetscErrorCode VectorField::sum
(
	PetscScalar *S
)
{
	PetscWrapper::ComputeSumOfPetscVectorElements(g_,S,n_);
	return(0);
}


PetscErrorCode ScalarField::arithmeticMean
(
	PetscScalar &m
)
{
	PetscWrapper::ComputeArithmeticMeanOfPetscVectorElements(g_,m);
	return(0);
}


PetscErrorCode VectorField::arithmeticMean
(
	PetscScalar *M
)
{
	PetscWrapper::ComputeArithmeticMeanOfPetscVectorElements(g_,M,n_);
	return(0);
}


PetscErrorCode VectorField::addTo
(
	ScalarField &f
)
{
	PetscWrapper::SumArrayOfPetscVectors(g_,f.g(),n_);
	return(0);
}


template<typename T>
PetscErrorCode GetArray
(
	T &f
)
{
	f.getArray();
	return(0);
}


template<typename T, typename ... Ts>
PetscErrorCode GetArray
(
	T& first,
	Ts& ... args
)
{
	first.getArray();
	GetArray(args ...);
	return(0);
}


template<typename T>
PetscErrorCode RestoreArray
(
	T &f
)
{
	f.restoreArray();
	return(0);
}


template<typename T, typename ... Ts>
PetscErrorCode RestoreArray
(
	T& first,
	Ts& ... args
)
{
	first.restoreArray();
	RestoreArray(args ...);
	return(0);
}


template PetscErrorCode GetArray<ScFi>(ScFi&);
template PetscErrorCode GetArray<VeFi>(VeFi&);
template PetscErrorCode GetArray<ScFi,ScFi>(ScFi&,ScFi&);
template PetscErrorCode GetArray<ScFi,VeFi>(ScFi&,VeFi&);
template PetscErrorCode GetArray<VeFi,ScFi>(VeFi&,ScFi&);
template PetscErrorCode GetArray<VeFi,VeFi>(VeFi&,VeFi&);
template PetscErrorCode GetArray<VeFi,ScFi,ScFi>(VeFi&,ScFi&,ScFi&);
template PetscErrorCode GetArray<ScFi,VeFi,ScFi>(ScFi&,VeFi&,ScFi&);
template PetscErrorCode GetArray<ScFi,ScFi,VeFi>(ScFi&,ScFi&,VeFi&);

template PetscErrorCode RestoreArray<ScFi>(ScFi&);
template PetscErrorCode RestoreArray<VeFi>(VeFi&);
template PetscErrorCode RestoreArray<ScFi,ScFi>(ScFi&,ScFi&);
template PetscErrorCode RestoreArray<ScFi,VeFi>(ScFi&,VeFi&);
template PetscErrorCode RestoreArray<VeFi,ScFi>(VeFi&,ScFi&);
template PetscErrorCode RestoreArray<VeFi,VeFi>(VeFi&,VeFi&);
template PetscErrorCode RestoreArray<VeFi,ScFi,ScFi>(VeFi&,ScFi&,ScFi&);
template PetscErrorCode RestoreArray<ScFi,VeFi,ScFi>(ScFi&,VeFi&,ScFi&);
template PetscErrorCode RestoreArray<ScFi,ScFi,VeFi>(ScFi&,ScFi&,VeFi&);

template PetscErrorCode GetArray<PetscWrapper::GridCoordinates3D,VeFi,ScFi>(PetscWrapper::GridCoordinates3D&,VeFi&,ScFi&);
template PetscErrorCode RestoreArray<PetscWrapper::GridCoordinates3D,VeFi,ScFi>(PetscWrapper::GridCoordinates3D&,VeFi&,ScFi&);


template<typename T>
PetscErrorCode GetGhostedArray
(
	T &f
)
{
	f.getGhostedArray();
	return(0);
}


template<typename T, typename ... Ts>
PetscErrorCode GetGhostedArray
(
	T& first,
	Ts& ... args
)
{
	first.getGhostedArray();
	GetGhostedArray(args ...);
	return(0);
}


template<typename T>
PetscErrorCode RestoreGhostedArray
(
	T &f
)
{
	f.restoreGhostedArray();
	return(0);
}


template<typename T, typename ... Ts>
PetscErrorCode RestoreGhostedArray
(
	T& first,
	Ts& ... args
)
{
	first.restoreGhostedArray();
	RestoreGhostedArray(args ...);
	return(0);
}


template PetscErrorCode GetGhostedArray<ScFi>(ScFi&);
template PetscErrorCode GetGhostedArray<VeFi>(VeFi&);
template PetscErrorCode GetGhostedArray<ScFi,ScFi>(ScFi&,ScFi&);
template PetscErrorCode RestoreGhostedArray<ScFi>(ScFi&);
template PetscErrorCode RestoreGhostedArray<VeFi>(VeFi&);
template PetscErrorCode RestoreGhostedArray<ScFi,ScFi>(ScFi&,ScFi&);
