#ifndef FITPACKWRAPPER_H
#define FITPACKWRAPPER_H


#include "petsc.h"
#include <Eigen/Dense>
#include <iostream>


/** \file
 *  \brief Wrapper to FitPack functions P. Dierckx used for splines */


/** \namespace FitPackWrapper
  * \brief Collection of FitPack functions */
namespace FitPackWrapper
{


	/** \brief Smoothing and resampling of an open parametric curve
	 *  \note The two extremities remain identical
	 *  \param[in]  mat      polyline in matrix form [X0, X1, X2, ...] with Xi = [xi,yi,zi]
	 *  \param[out] mats     smoothed polyline in matrix form [X0, X1, X2, ...] with Xi = [xi,yi,zi]
	 *  \param[in]  sf       smoothing factor
	 *  \param[in]  isClosed closed or open polyline flag
	 *  \param[out] error    error flag */
	int SmoothPolyline
	(
		const Eigen::MatrixXf &mat,
		Eigen::MatrixXf &mats,
		float sf,
		bool isClosed,
		bool &error
	);


	/** @copydoc FitPackWrapper::SmoothPolyline(const Eigen::MatrixXf&,Eigen::MatrixXf&,float,bool,bool&) */
	template<typename Type>
	int SmoothPolyline
	(
		const Eigen::Matrix<Type,Eigen::Dynamic,Eigen::Dynamic> &mat,
		Eigen::Matrix<Type,Eigen::Dynamic,Eigen::Dynamic> &mats,
		Type sf,
		bool isClosed,
		bool &error
	);


}


#endif
