#include "Hinge.h"
#include <fstream>
#include <iostream>


Bead::Bead()
{}


// Bead::Bead():
// 	density_{1.},
// 	radius_{1.},
// 	mass_{1.},
// 	inertia_{1.},
// 	position_{EigenWrapper::EigenVec3::Zero()},
// 	quaternion_{EigenWrapper::EigenQuat::Identity()},
// 	vel_{EigenWrapper::EigenVec3::Zero()},
// 	omg_{EigenWrapper::EigenVec3::Zero()},
// 	force_{EigenWrapper::EigenVec3::Zero()},
// 	torque_{EigenWrapper::EigenVec3::Zero()},
// 	forceOld_{EigenWrapper::EigenVec3::Zero()},
// 	torqueOld_{EigenWrapper::EigenVec3::Zero()}
// {}


Bead::Bead
(
	const Bead &rhs
):
	density_{rhs.density_},
	radius_{rhs.radius_},
	mass_{rhs.mass_},
	inertia_{rhs.inertia_},
	position_{rhs.position_},
	quaternion_{rhs.quaternion_},
	vel_{rhs.vel_},
	omg_{rhs.omg_},
	force_{rhs.force_},
	torque_{rhs.torque_},
	forceOld_{rhs.forceOld_},
	torqueOld_{rhs.torqueOld_}
{}


Bead::~Bead()
{}


Bead& Bead::operator=
(
	const Bead& rhs
)
{
	if ( this != &rhs )
	{
		density_ = rhs.density_;
		radius_ = rhs.radius_;
		mass_ = rhs.mass_;
		inertia_ = rhs.inertia_;
		position_ = rhs.position_;
		quaternion_ = rhs.quaternion_;
		vel_ = rhs.vel_;
		omg_ = rhs.omg_;
		force_ = rhs.force_;
		torque_ = rhs.torque_;
		forceOld_ = rhs.forceOld_;
		torqueOld_ = rhs.torqueOld_;
	}
	return *this;
}


PetscErrorCode Bead::setProperties
(
	PetscScalar density,
	PetscScalar radius
)
{
	density_ = density;
	radius_ = radius;
	mass_ = density_ * (4./3.) *(PETSC_PI) * PetscPowRealInt(radius_,3);
	inertia_ = (2./5.) * mass_* PetscSqr(radius_);
	return 0;
}


Hinge::Hinge
(
	const std::string& name
):
	name_{name}
{}


Hinge::Hinge
(
	PetscScalar density,
	const EigenWrapper::EigenVec3& A,
	const EigenWrapper::EigenVec3& B,
	PetscInt n,
	const std::string& name
):
	name_{name}
{
	this->set(density,A,B,n);
}


Hinge::Hinge
(
	const EigenWrapper::EigenVec3& A,
	const EigenWrapper::EigenVec3& B,
	PetscInt n,
	const std::string& name
):
	name_{name}
{
	this->set(1.,A,B,n);
}


Hinge::Hinge()
{}


Hinge::Hinge
(
	const Hinge& rhs
):
	name_{rhs.name_},
	density_{rhs.density_},
	radius_{rhs.radius_},
	length_{rhs.length_},
	ks_{rhs.ks_},
	damping_{rhs.damping_},
	beads_{rhs.beads_},
	size_{rhs.size_}
{}


Hinge& Hinge::operator=
(
	const Hinge& rhs
)
{
	if (this != &rhs)
	{
		name_    = rhs.name_;
		density_ = rhs.density_;
		radius_  = rhs.radius_;
		length_  = rhs.length_;
		ks_      = rhs.ks_;
		damping_ = rhs.damping_;
		beads_   = rhs.beads_;
		size_    = rhs.size_;
	}
	return *this;
}


Hinge::~Hinge()
{}


PetscErrorCode Hinge::set
(
	PetscScalar density,
	const EigenWrapper::EigenVec3& A,
	const EigenWrapper::EigenVec3& B,
	PetscInt size
)
{
	size_ = size;
	beads_ = std::vector<Bead>(size_);
	for (PetscInt b=0; b<size_; ++b)
	{
		PetscScalar delta = static_cast<PetscScalar>(b)/static_cast<PetscScalar>(size_-1);
		beads_[b].position() = A + delta * (B-A);
		//PetscPrintf(PETSC_COMM_WORLD,"  P%i = %4.2f %4.2f %4.2f\n",b,beads_[b].position().x(),beads_[b].position().y(),beads_[b].position().z());
	}
	PetscScalar radius = (beads_[1].position() - beads_[0].position()).norm()/2.;
	this->setProperties(density,radius);
	const auto& b = beads_[0];
	PetscPrintf(PETSC_COMM_WORLD,"  %s set\n",name_.c_str());
	PetscPrintf(PETSC_COMM_WORLD,"    rho = %4.4f r = %4.4f m = %4.4f I = %4.4f\n",b.density(),b.radius(),b.mass(),b.inertia());
	return(0);
}


PetscErrorCode Hinge::setProperties
(
	PetscScalar density,
	PetscScalar radius
)
{
	density_ = density;
	radius_ = radius;
	length_ = 2.*(size_-1)*(radius_);
	for (auto &bead : beads_)
	{
		bead.setProperties(density,radius);
	}
	return(0);
}


PetscErrorCode Hinge::writeInASCIIFormat
(
	PetscInt t
) const
{
	EigenWrapper::EigenMat m(size_,23);
	for (PetscInt b=0; b<size_; ++b)
	{
		EigenWrapper::EigenVec v(m.cols());
		v <<
			beads_[b].density(),             //1
			beads_[b].radius(),              //1
			beads_[b].mass(),                //1
			beads_[b].inertia(),             //1
			beads_[b].position(),            //3
			beads_[b].quaternion().coeffs(), //4
			beads_[b].vel(),                 //3
			beads_[b].omg(),                 //3
			beads_[b].force(),               //3
			beads_[b].torque();              //3
		m.row(b) = v.transpose();
	}
	PetscInt rank; MPI_Comm_rank(PETSC_COMM_WORLD,&rank);
	PetscInt source = 0;
	if (rank==source)
	{
		EigenWrapper::WriteInASCIIFormat(m,name_,t);
	}
	return(0);
}


PetscErrorCode Hinge::writeInBinaryFormat
(
	PetscInt t
) const
{
	char fname[1000];
	if (t<0)
	{
		sprintf(fname,"%s.dat",name_.c_str());
	}
	else
	{
		sprintf(fname,"%s_t%05d.dat",name_.c_str(),t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Write %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_WRITE,&viewer);
	PetscViewerBinaryWrite(viewer,&density_,1,PETSC_SCALAR);
	PetscViewerBinaryWrite(viewer,&radius_,1,PETSC_SCALAR);
	PetscViewerBinaryWrite(viewer,&length_,1,PETSC_SCALAR);
	PetscViewerBinaryWrite(viewer,&ks_,1,PETSC_SCALAR);
	PetscViewerBinaryWrite(viewer,&damping_,1,PETSC_SCALAR);
	PetscViewerBinaryWrite(viewer,&size_,1,PETSC_INT);
	EigenWrapper::EigenMat m(size_,23);
	for (PetscInt b=0; b<size_; ++b)
	{
		const auto& bead = beads_[b];
		EigenWrapper::EigenVec v(m.cols());
		v <<
			bead.density(),             //1
			bead.radius(),              //1
			bead.mass(),                //1
			bead.inertia(),             //1
			bead.position(),            //3
			bead.quaternion().coeffs(), //4
			bead.vel(),                 //3
			bead.omg(),                 //3
			bead.force(),               //3
			bead.torque();              //3
		m.row(b) = v;
	}
	PetscViewerBinaryWrite(viewer,m.data(),m.size(),PETSC_SCALAR);
	PetscViewerDestroy(&viewer);
	return(0);
}


PetscErrorCode Hinge::readInBinaryFormat
(
	PetscInt t
)
{
	char fname[1000];
	if (t<0)
	{
		sprintf(fname,"%s.dat",name_.c_str());
	}
	else
	{
		sprintf(fname,"%s_t%05d.dat",name_.c_str(),t);
	}
	PetscPrintf(PETSC_COMM_WORLD,"  Read %s\n",fname);
	PetscViewer viewer;
	PetscViewerBinaryOpen(PETSC_COMM_WORLD,fname,FILE_MODE_READ,&viewer);
	PetscViewerBinaryRead(viewer,&density_,1,NULL,PETSC_SCALAR); MPI_Bcast(&density_,1,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	PetscViewerBinaryRead(viewer,&radius_ ,1,NULL,PETSC_SCALAR); MPI_Bcast(&radius_ ,1,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	PetscViewerBinaryRead(viewer,&length_ ,1,NULL,PETSC_SCALAR); MPI_Bcast(&length_ ,1,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	PetscViewerBinaryRead(viewer,&ks_     ,1,NULL,PETSC_SCALAR); MPI_Bcast(&ks_     ,1,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	PetscViewerBinaryRead(viewer,&damping_,1,NULL,PETSC_SCALAR); MPI_Bcast(&damping_,1,MPIU_SCALAR,0,PETSC_COMM_WORLD);
	PetscViewerBinaryRead(viewer,&size_   ,1,NULL,PETSC_INT   ); MPI_Bcast(&size_   ,1,MPIU_INT   ,0,PETSC_COMM_WORLD);
	beads_.resize(size_);
	EigenWrapper::EigenMat m(size_,23);
	PetscViewerBinaryRead(viewer,m.data(),m.size(),NULL,PETSC_SCALAR);
	MPI_Bcast(m.data(),m.size(),MPIU_SCALAR,0,PETSC_COMM_WORLD);
	for (PetscInt b=0; b<size_; ++b)
	{
		auto& bead = beads_[b];
		const EigenWrapper::EigenVec& v = m.row(b);
		bead.density()             = v(0);
		bead.radius()              = v(1);
		bead.mass()                = v(2);
		bead.inertia()             = v(3);
		bead.position()            = v.segment( 4,3);
		bead.quaternion().coeffs() = v.segment( 7,4);
		bead.vel()                 = v.segment(11,3);
		bead.omg()                 = v.segment(14,3);
		bead.force()               = v.segment(17,3);
		bead.torque()              = v.segment(20,3);
		bead.forceOld()            = bead.force();
		bead.torqueOld()           = bead.torque();
	}
	PetscViewerDestroy(&viewer);
	return 0;
}


PetscErrorCode Hinge::updateBeadForces()
{
	//Local constants
	const EigenWrapper::EigenVec3 ex(1,0,0);
	const EigenWrapper::EigenVec3 ey(0,1,0);
	const EigenWrapper::EigenVec3 ez(0,0,1);
	const PetscScalar ksR = (3./4.) * (ks_) * (radius_);
	const PetscScalar ksR2 = (1./2.) * (ks_) * PetscSqr(radius_) ;

	//Reset beam forces/torques to zero
	for (auto &bead:beads_)
	{
		bead.forceOld() = bead.force();
		bead.torqueOld() = bead.torque();
		bead.force().setZero();
		bead.torque().setZero();
	}

	//Compute beam forces/torques
	//Loop over virtual beams, that is from 0 to n-2
	for (PetscInt b=0; b<(size_-1); ++b)
	{
		//bi,bj
		auto& beadi = beads_[b  ];
		auto& beadj = beads_[b+1];

		//Unit bond vectors bx,by,bz
		EigenWrapper::EigenVec3 bx,by,bz;
		bx = beadj.position() - beadi.position();
		PetscScalar dist = bx.norm();
		bx /= dist;

		//Quaternion associated with frame transformation (x,y,z) -> (x',y',z') with x=(1,0,0), x'=(bx,by,bz)
		EigenWrapper::EigenQuat qFrame;
		qFrame.setFromTwoVectors(ex,bx);
		qFrame.normalize();

		//Unit shear vectors by,bz
		by = qFrame.toRotationMatrix() * ey;
		bz = qFrame.toRotationMatrix() * ez;

		//Substract frame rotation
		EigenWrapper::EigenQuat qInvFrame = qFrame.conjugate();
		EigenWrapper::EigenQuat qi = ( beadi.quaternion() ) * qInvFrame;
		EigenWrapper::EigenQuat qj = ( beadj.quaternion() ) * qInvFrame;

		//Orientation angles in beam frame
 		//PetscScalar thetax_i = 2.*PetscAsinReal(qi.x()); PetscScalar thetaj_x = 2.*PetscAsinReal(qj.x());
		PetscScalar thetay_i = 2.*PetscAsinReal(qi.y()); PetscScalar thetay_j = 2.*PetscAsinReal(qj.y());
		PetscScalar thetaz_i = 2.*PetscAsinReal(qi.z()); PetscScalar thetaz_j = 2.*PetscAsinReal(qj.z());

		//Normal force in x-direction
		EigenWrapper::EigenVec3 Fx = (+ks_) * ( dist - 2.*radius_ ) * bx;
		beadi.force() += Fx;
		beadj.force() -= Fx;

		//Shear force in y-direction
		EigenWrapper::EigenVec3 Fy = (-ksR)*(thetaz_i + thetaz_j) * by;
		beadi.force() += Fy;
		beadj.force() -= Fy;

		//Shear force in z-direction
		EigenWrapper::EigenVec3 Fz = (+ksR)*(thetay_i + thetay_j) * bz;
		beadi.force() += Fz;
		beadj.force() -= Fz;

		//Torsion torque in x-direction
		// Ignored

		//Bending torques y-direction
		EigenWrapper::EigenVec3 Ty_i = (-ksR2) * (2.*thetay_i +    thetay_j) * by;
		EigenWrapper::EigenVec3 Ty_j = (-ksR2) * (   thetay_i + 2.*thetay_j) * by;
		beadi.torque() += Ty_i;
		beadj.torque() += Ty_j;

		//Bending torques z-direction
		EigenWrapper::EigenVec3 Tz_i = (-ksR2) * (2.*thetaz_i +    thetaz_j) * bz;
		EigenWrapper::EigenVec3 Tz_j = (-ksR2) * (   thetaz_i + 2.*thetaz_j) * bz;
		beadi.torque() += Tz_i;
		beadj.torque() += Tz_j;
	}

	//Material damping
	for (auto &bead:beads_)
	{
		bead.force()  -= (damping_) * bead.vel() /* * bead.vel().norm() */;
		bead.torque() -= (damping_) * bead.omg() /* * bead.omg().norm() */;
	}

	return(0);
}


PetscErrorCode Hinge::updateBeadVelocities
(
	PetscScalar dt
)
{
	for (auto& bead:beads_)
	{
		//Integrate translational velocity
// 		EigenWrapper::EigenVec3 force = bead.force();
		EigenWrapper::EigenVec3 force = 1.5 * bead.force() - 0.5 * bead.forceOld();
		bead.vel() += force * dt/bead.mass();

		//Integrate rotational velocity (non-spherical beads)
// 		EigenWrapper::EigenMat3 rotationMatrix = bead.quaternion().toRotationMatrix();
// 		EigenWrapper::EigenMat3 rotationMatrixT = rotationMatrix.transpose();
// 		EigenWrapper::EigenVec3 omgHat  = rotationMatrixT * bead.omg();
// 		EigenWrapper::EigenVec3 torqueHat = rotationMatrixT * bead.torque();
// 		omgHat += torqueHat * (dt)/( bead.inertia() );
// 		bead.omg() = rotationMatrix * omgHat;

		//Integrate rotational velocity (spherical beads)
		EigenWrapper::EigenVec3 torque = bead.torque();
// 		EigenWrapper::EigenVec3 torque = 1.5 * bead.torque() - 0.5 * bead.torqueOld(); //ERROR
		bead.omg() += torque * dt/bead.inertia();
	}
	return(0);
}


PetscErrorCode Hinge::updateBeadPositions
(
	PetscScalar dt
)
{
	for (auto &bead:beads_)
	{
		//Integrate bead position
		bead.position() += bead.vel() * dt;

		//Integrate bead quaternion
		const PetscScalar Ox = bead.omg().x();
		const PetscScalar Oy = bead.omg().y();
		const PetscScalar Oz = bead.omg().z();
		const EigenWrapper::EigenQuat qold = bead.quaternion();
		const PetscScalar qw = qold.w();
		const PetscScalar qx = qold.x();
		const PetscScalar qy = qold.y();
		const PetscScalar qz = qold.z();
		EigenWrapper::EigenQuat qnew;
		qnew.w() = qw + (dt/2.)*(        - Ox*qx - Oy*qy - Oz*qz);
		qnew.x() = qx + (dt/2.)*(+ Ox*qw         - Oz*qy + Oy*qz);
		qnew.y() = qy + (dt/2.)*(+ Oy*qw + Oz*qx         - Ox*qz);
		qnew.z() = qz + (dt/2.)*(+ Oz*qw - Oy*qx + Ox*qy        );
		qnew.normalize();
		bead.quaternion() = qnew;
	}
	return(0);
}


PetscErrorCode Hinge::update
(
	const EigenWrapper::EigenVec3 &Fext,
	const EigenWrapper::EigenVec3 &Text,
	PetscScalar dt
)
{
	this->updateBeadForces();
	beads_.front().force().setZero();
	beads_.front().torque().setZero();
	beads_.back().force()  += Fext;
	beads_.back().torque() += Text;
	this->updateBeadVelocities(dt);
	this->updateBeadPositions(dt);
	return(0);
}


PetscScalar Hinge::referenceForce() const
{
	return 0.5* ks_ * PetscPowRealInt(radius_,3) / PetscSqr(length_);
}


PetscScalar Hinge::normalizedLoad
(
	PetscScalar load
) const
{
	return load / this->referenceForce();
}


EigenWrapper::EigenVec3 Hinge::normalizedLoad
(
	const EigenWrapper::EigenVec3& load
) const
{
	return load / this->referenceForce();
}


PetscErrorCode Hinge::check() const
{
	for (const auto& bead : beads_)
	{
		if ( radius_ != bead.radius() )
		{
			PetscPrintf(PETSC_COMM_WORLD,"  Bead radius != hinge radius\n");
			PetscEnd();
		}
		if
		(
			PetscIsInfOrNanScalar(bead.position().x()) ||
			PetscIsInfOrNanScalar(bead.position().y()) ||
			PetscIsInfOrNanScalar(bead.position().z()) ||
			bead.position().norm() > 1e6
		)
		{
			PetscPrintf(PETSC_COMM_WORLD,"  Bead position -> infinity\n");
			PetscEnd();
		}
	}
	return(0);
}


PetscErrorCode TestHinge
(
	PetscScalar load
)
{
	PetscScalar dt = 0.1;
	PetscInt tmax = 5000;
	PetscInt save = 10;
	PetscScalar density = 1.;
	PetscInt nb = 10;
	EigenWrapper::EigenVec3 A(0,0,0);
	EigenWrapper::EigenVec3 B(4,0,0);
	Hinge h = Hinge(density,A,B,nb,"hinge");
// 	for (auto&bead:h.beads())
// 	{
// 		bead.mass() = 1;
// 		bead.inertia() = 1;
// 	}
	h.damping() = 0.003;
	PetscScalar nomi = ( h.ks() ) * PetscPowRealInt(h.radius(),3);
	PetscScalar denomi = 2. * PetscSqr(h.length());
	PetscScalar Fy = load * (nomi/denomi);
	for (PetscInt t=0; t<tmax; ++t)
	{
		h.update
		(
			EigenWrapper::EigenVec3(0,Fy,0),
			EigenWrapper::EigenVec3::Zero(),
			dt
		);
		h.check();
		if (t % save == 0)
		{
			h.writeInASCIIFormat(t);
		}
	}
	PetscPrintf(PETSC_COMM_WORLD,"HingeToVTK.py -hg hinge -t 0:5000:10\n");
	return(0);
}
