#include "EigenWrapper.h"
#include "Polyline.h"
#include "igl/read_triangle_mesh.h"
#include "igl/writeSTL.h"

static char help[] = "Inspect polylines obtained upon warning\n\n";


PetscErrorCode PrintUsage()
{
	std::cout
		<< "  USAGE:" << std::endl
		<< "  -h  " << " print usage" << std::endl
		<< "  -t  " << " time index at which error occurred" << std::endl
		<< "  -np " << " number of procs used in the simulation" << std::endl
		<< "  -eps" << " distance used in smoothing operation" << std::endl
		<< "  -s  " << " smoothing factor" << std::endl
		<< "  Inspect polylines obtained upon warning" << std::endl
	;
	return(0);
}

PetscInt main(int argc,char* argv[])
{
	PetscInitialize(&argc,&argv,(char*)0,help);
	PetscBool usage(PETSC_FALSE); PetscOptionsGetBool(NULL,NULL,"-h",&usage,NULL);
	if (usage)
	{
		PrintUsage();
		PetscEnd();
	}
	PetscBool flg;
	PetscInt t(0); PetscOptionsGetInt(NULL,NULL,"-t",&t,&flg);
	PetscInt np(16); PetscOptionsGetInt(NULL,NULL,"-np",&np,&flg);
	PetscScalar eps(0.1); PetscOptionsGetScalar(NULL,NULL, "-eps", &eps,&flg);
	PetscScalar s(1); PetscOptionsGetScalar(NULL,NULL, "-s", &s,&flg);
	std::string name;
	EigenWrapper::EigenMat wallVertices,phiVertices;
	EigenWrapper::EigenMatI wallFaces,phiFaces;
	name = "err_it" + std::to_string(t) + "_wall";
	std::cout << "  Read " << name << ".off" << std::endl;
	igl::read_triangle_mesh(name + ".off",wallVertices,wallFaces);
	igl::writeSTL(name + ".stl",wallVertices,wallFaces);
	std::vector<Polyline> v;
	for (PetscInt rank=0; rank<np; ++rank)
	{
		phiVertices.resize(0,0);
		phiFaces.resize(0,0);
		name = "err_rk" + std::to_string(rank) + "_it" + std::to_string(t) + "_phi";
		std::cout << "  Read " << name << ".off -> ";
		igl::read_triangle_mesh(name + ".off",phiVertices,phiFaces);
		igl::writeSTL(name + ".stl",phiVertices,phiFaces);
		std::vector<Polyline> localv = Polylines(phiVertices,phiFaces,wallVertices,wallFaces).pls();
		std::cout << localv.size() << " polyline(s) found" << std::endl;
		v.insert(v.end(),localv.begin(),localv.end());
	}
	Polylines inpect(v,"inspect");
	inpect.concatenateAll();
	inpect.cleanAll(eps);
	if (s>0)
	{
		inpect.smoothAll(eps,s);
	}
	inpect.writeInBinaryFormat(-1);
	PetscFinalize();
	return(0);
}
