#include "Context3D.h"
#include "Constant.h"

#include "SolidBody.h"

// export IGL_NUM_THREADS=1
static char help[] = "Microorigami\n\n";

PetscInt main(int argc,char* argv[])
{
	PetscInitialize(&argc,&argv,(char*)0,help);
	Constant::ReadInputFile();
	DM da;
	DMDACreate3d
	(
		PETSC_COMM_WORLD,     //MPI communicator
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the x-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the y-direction
		DM_BOUNDARY_PERIODIC, //type of ghost nodes the array has in the z-direction
		DMDA_STENCIL_BOX,     //stencil type (DMDA_STENCIL_BOX or DMDA_STENCIL_STAR)
		(Constant::Grid::Nx), //global dimension in x-direction of the array
		(Constant::Grid::Ny), //global dimension in y-direction of the array
		(Constant::Grid::Nz), //global dimension in y-direction of the array
		PETSC_DECIDE,         //corresponding number of processors in x-dimension
		PETSC_DECIDE,         //corresponding number of processors in y-dimension
		PETSC_DECIDE,         //corresponding number of processors in z-dimension
		1,                    //number of degrees of freedom per node (i.e. number of field components)
		1,                    //stencil width (1, 2 or 3), 3 for WENO
		NULL,                 //arrays containing the number of nodes in each cell along the x coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the y coordinates
		NULL,                 //arrays containing the number of nodes in each cell along the z coordinates
		&da                   //dynamic array
	);
	DMSetUp(da);
	DMDASetUniformCoordinates
	(
		da,
		0.,(Constant::Grid::L[0]),
		0.,(Constant::Grid::L[1]),
		0.,(Constant::Grid::L[2])
	);
	PetscInt n = (Constant::Template::Sides);

// 	//Fig5: Hinge model validation
// 	{
// 		Context ctx
// 		(
// 			da,
// 			Constant::Template::Sides
// 		);
// 		PetscScalar Lx = Constant::Grid::L[0];
// 		PetscScalar baseLength = Lx/10.;
// 		PetscScalar baseArea = baseLength*baseLength;
// 		PetscScalar pyramidHeight = Lx/3;
// 		ctx.initialize
// 		(
// 			baseArea,
// 			pyramidHeight
// 		);
// 		ctx.validateHingeModel();
// 		// tend=20000; every=200; s=0; loadMax=31;
// 		// for (( load=0; load<loadMax; load++ )); do ExportHinge -f hinge0_load${load} -t 0:${tend}:${every}; done
// 		// HingeToVTK.py -f hinge${s}_load1 -t 0:${tend}:${every}
// 		// ExportSolid -f base -stl
// 		// ExportSolid -f side${s}_load1 -tp -stl -t 0:${tend}:${every}
// 		// load=21; PlotFibre.py -f hinge0_load${load} -t 0:5000:100
// 	}

	//Fig6: Capillary model validation
// 	if (n<1)
// 	{
// 		Context ctx(da,-1);
// 		ctx.initialize();
// 		ctx.validateCapillaryForceModel();
// 	}

	//Final
	{
		PetscScalar Lx = Constant::Grid::L[0];
		PetscScalar baseLength = Lx/10.;
		PetscScalar pyramidHeight = Lx/3;
		PetscScalar baseArea = baseLength*baseLength;
		Context ctx(da,Constant::Template::Sides);
		ctx.initialize(baseArea,pyramidHeight);
		ctx.equilibriate();
		ctx.run();

		// tend=100; every=1; side=6;
		// ExportSolid -f base -stl;
		// for (( s=0; s<${side}; s++ )); do ExportSolid -f side${s} -t 0:${tend}:${every} -tp -stl -txt; done;
		// for (( h=0; h<${side}; h++ )); do ExportHinge -f hinge${h} -t 0:${tend}:${every}; done
		// HingeToVTK.py -f hinge -n ${side} -t 0:${tend}:${every};
		// for f in hinge?_t?????.py; do echo $f; pvpython $f; done;
		// ExportFieldToVTS -f phi -t 0:${tend}:${every};
		// ExportPolylines -f tpcl -t 0:${tend}:${every};
		// PolylineToVTK.py -f tpcl -n 1 -t 0:${tend}:${every};
		// for f in tpcl*.py; do pvpython $f; done
	}

	DMDestroy(&da);
	PetscFinalize();
	return(0);
}

// make -j 6 main post
// tend=20; every=1; ExportSolid -i Input.in -s base -stl; ExportSolid -i Input.in -s side0 -stl -n 3 -t 0:${tend}:${every}; ExportFieldToVTS -i Input.in -f phi -t 0:${tend}:${every}; ExportPolylines -c tpcl -t 0:${tend}:${every}; PolylineToVTK.py -c tpcl -n 1 -t 0:${tend}:${every}; for f in tpcl??_t?????.py; do pvpython $f; done




	//Force folding
// 	{
// 		PetscScalar Lx = Constant::Grid::L[0];
// 		PetscScalar baseLength = Lx/10.;
// 		PetscScalar pyramidHeight = Lx/3;
// 		PetscScalar baseArea = baseLength*baseLength;
// 		Context ctx
// 		(
// 			da,
// 			Constant::Template::Sides
// 		);
// 		ctx.initialize(baseArea,pyramidHeight);
// 		ctx.fold();
// /*
// 		rm -f *_t?????.stl *_t?????.vts *_t?????.vtk *_t?????.txt *_t?????.py
// 		tend=11; every=1; side=4;
// 		ExportSolid -f base -stl;
// 		for (( s=0; s<${side}; s++ )); do ExportSolid -f side${s} -t 0:${tend}:${every} -tp -stl; done;
// 		ExportFieldToVTS -f phi -t 0:${tend}:${every};
// 		ExportPolylines -f tpcl -t 0:${tend}:${every};
// 		PolylineToVTK.py -f tpcl -n 3 -t 0:${tend}:${every};
// 		for f in tpcl*.py; do pvpython $f; done
// */
// 	}

	//Polylines
// 	{
// 		PetscScalar Lx = Constant::Grid::L[0];
// 		PetscScalar baseLength = Lx/10.;
// 		PetscScalar pyramidHeight = Lx/3;
// 		PetscScalar baseArea = baseLength*baseLength;
// 		Context ctx(da,Constant::Template::Sides);
// 		ctx.initialize(baseArea,pyramidHeight);
// 		ctx.equilibriate();
// 		ctx.updateContactLine(1);
//
// 		// side=3
// 		// ExportFieldToVTS -f phi -t 1;
// 		// ExportSolid -f base -stl;
// 		// for (( s=0; s<${side}; s++ )); do ExportSolid -f side${s} -t 1 -tp -stl -txt; done;
// 		// for (( h=0; h<${side}; h++ )); do ExportHinge -f hinge${h} -t 1; done;
// 		// HingeToVTK.py -f hinge -n 3 -t 1 -a 12
// 		// for f in hinge?_t?????.py; do echo $f; pvpython $f; done;
// 		// cp err_it1_tpcl_con00.txt tpcl_rough.txt
// 		// sed -n '1~2p' err_it1_tpcl_smo00.txt > tpcl_smooth.txt
// 		// PolylineToVTK.py -f tpcl_rough.txt; pvpython tpcl_rough.py
// 		// PolylineToVTK.py -f tpcl_smooth.txt; pvpython tpcl_smooth.py
// 	}
