#include "SNESWrapper.h"


PetscErrorCode SNESWrapper::CreatePetscNonLinearSolver
(
	DM da,
	PetscNonLinearSolver &solver,
	const std::string& name
)
{
	solver.name = name;
	SNESCreate(PETSC_COMM_WORLD,&solver.snes);
	DMCreateMatrix(da,&solver.jac);
	DMCreateGlobalVector(da,&solver.solution);
	DMCreateGlobalVector(da,&solver.residual);
	SNESSetDM(solver.snes,da);
	SNESSetFromOptions(solver.snes);
	PetscPrintf(PETSC_COMM_WORLD,"  SNES object %s created \n", solver.name.c_str());
	return(0);
}


PetscErrorCode SNESWrapper::DestroyPetscNonLinearSolver
(
	PetscNonLinearSolver &solver
)
{
	SNESDestroy(&solver.snes);
	MatDestroy(&solver.jac);
	VecDestroy(&solver.solution);
	VecDestroy(&solver.residual);
	PetscPrintf(PETSC_COMM_WORLD,"  SNES solver %s destroyed \n", solver.name.c_str());
	return(0);
}


PetscErrorCode SNESWrapper::RunPetscNonLinearSolver
(
	PetscNonLinearSolver &solver,
	PetscErrorCode(*SNESFunctionPtr)(SNES,Vec,Vec,void*),
	PetscErrorCode(*SNESJacobianPtr)(SNES,Vec,Mat,Mat,void*),
	void *ctxPtr
)
{
	SNESSetFunction
	(
		solver.snes,
		solver.residual,
		SNESFunctionPtr,
		ctxPtr
	);
	SNESSetJacobian
	(
		solver.snes,
		solver.jac,
		solver.jac,
		SNESJacobianPtr,
		ctxPtr
	);
	SNESSolve
	(
		solver.snes,
		NULL,
		solver.solution
	);
// 	PetscPrintf(PETSC_COMM_WORLD,"  %s equation solved\n",solver.name.c_str());
	return(0);
}


PetscErrorCode SNESWrapper::SetNonLinearSolverSolution
(
	PetscNonLinearSolver &solver,
	Vec sol
)
{
	VecCopy(sol,solver.solution);
	return(0);
}


PetscErrorCode SNESWrapper::GetNonLinearSolverSolution
(
	PetscNonLinearSolver &solver,
	Vec sol
)
{
	VecCopy(solver.solution,sol);
	return(0);
}