#ifndef UTILS_H
#define UTILS_H


/** \file
  * \brief Useful macros */

/** \def GetDirectionIndices()
    \brief %Constant integers associated with x, y, z, centre, north, south, east, west, front, back */
#define GetDirectionIndices() \
namespace \
{ \
	const PetscInt X_ = 0; /**< \brief Space direction = 0 */ \
	const PetscInt Y_ = 1; /**< \brief Space direction = 1 */ \
	const PetscInt Z_ = 2; /**< \brief Space direction = 2 */ \
	const PetscInt C_ = 0; /**< \brief Centre index = 0 */ \
	const PetscInt F_ = 1; /**< \brief Front index = 1 */ \
	const PetscInt B_ = 2; /**< \brief Back index = 2 */ \
	const PetscInt N_ = 3; /**< \brief North index = 3 */ \
	const PetscInt S_ = 4; /**< \brief South index = 4 */ \
	const PetscInt E_ = 5; /**< \brief East index = 5 */ \
	const PetscInt W_ = 6; /**< \brief West index = 6 */ \
}


/** \def ForAllKJI(da,...)
    \brief Loop over indices in the 3D domain
    \note  https://mort.coffee/home/obscure-c-features/ */
#define ForAllKJI(da,...) \
{\
	PetscInt i0,j0,k0,iw,jw,kw;\
	DMDAGetCorners(da,&i0,&j0,&k0,&iw,&jw,&kw);\
	for (PetscInt k=k0; k<k0+kw; ++k){\
	for (PetscInt j=j0; j<j0+jw; ++j){\
	for (PetscInt i=i0; i<i0+iw; ++i){\
	__VA_ARGS__ }}}\
}


/** \def ForAllDim
    \brief Loop over the dimensions */
#define ForAllDim()\
	for (PetscInt dim=0; dim<3; ++dim)


// #define ForAllJI(da)
// 	PetscInt i0,j0,iw,jw;
// 	DMDAGetCorners(da,&i0,&j0,NULL,&iw,&jw,NULL);
// 	for (PetscInt j=j0; j<j0+jw; ++j)
// 	for (PetscInt i=i0; i<i0+iw; ++i)


#endif
